import datetime as _datetime
import json
import linecache
import sys
import typing as t
from dataclasses import dataclass
from pathlib import Path
from re import search

import dill
import msgspec
import toml
import yaml
from anyio import Path as AsyncPath

__all__: t.List[str] = ["load", "dump", "encode", "decode"]


def yaml_encode(
    obj: t.Any,
    *,
    enc_hook: t.Callable[[t.Any], t.Any] | None = None,
    sort_keys: bool = False,
) -> bytes:
    dumper = getattr(yaml, "CSafeDumper", yaml.SafeDumper)
    dumper.add_representer(
        type(None),
        lambda dumper, value: dumper.represent_scalar("tag:yaml.org,2002:null", ""),
    )
    return yaml.dump_all(
        [
            msgspec.yaml._to_builtins(
                obj,
                builtin_types=(_datetime.datetime, _datetime.date),
                enc_hook=enc_hook,
            )
        ],
        encoding="utf-8",
        Dumper=dumper,  # type: ignore
        allow_unicode=True,
        sort_keys=sort_keys,
    )


msgspec.yaml.encode = yaml_encode


@dataclass
class Serializers:
    json: t.Any = msgspec.json
    yaml: t.Any = msgspec.yaml
    msgpack: t.Any = msgspec.msgpack
    pickle: t.Any = dill
    toml: t.Any = msgspec.toml

    def __post_init__(self) -> None:
        self.pickle.encode = dill.dumps
        self.pickle.decode = dill.loads


serializers = Serializers()


@t.runtime_checkable
class SerializerMethod(t.Protocol):
    async def __call__(
        self,
        obj: t.Any,
        path: t.Optional[AsyncPath | Path] = None,
        sort_keys: bool = False,
        **kwargs: t.Any,
    ) -> t.Any: ...


class Encode:
    serializers: t.Dict[str, t.Any]
    path: t.Optional[AsyncPath | Path] = None
    action: t.Optional[str] = None
    sort_keys: bool = True
    use_list: bool = False
    secure: bool = False
    secret_key: t.Optional[str] = None
    secure_salt: t.Optional[str] = None
    serializer: t.Optional[t.Callable[..., t.Any]] = None

    json: SerializerMethod
    yaml: SerializerMethod
    msgpack: SerializerMethod
    pickle: SerializerMethod
    toml: SerializerMethod

    def __init__(self) -> None:
        self.serializers = serializers.__dict__
        for s in self.serializers:
            setattr(self, s, self._create_method(s, "encode"))

    def _create_method(
        self, serializer_name: str, default_action: str
    ) -> t.Callable[..., t.Any]:
        async def method(
            obj: t.Any,
            path: AsyncPath | Path | None = None,
            sort_keys: bool = False,
            **kwargs: t.Dict[str, t.Any],
        ) -> t.Any:
            self.path = path
            self.sort_keys = sort_keys

            frame = sys._getframe(1)
            caller_name = frame.f_code.co_name

            if caller_name.startswith("test_"):
                if "decode" in caller_name or default_action == "decode":
                    self.action = "decode"
                else:
                    self.action = "encode"
            else:
                self.action = default_action

            if serializer_name in self.serializers:
                serializer = self.serializers[serializer_name]

                if self.action == "encode":
                    self.serializer = getattr(serializer, "encode", None)
                else:
                    self.serializer = getattr(serializer, "decode", None)

                if self.serializer is None:
                    raise AttributeError(
                        f"No {self.action} method found for {serializer_name}"
                    )

                return await self.process(obj, **kwargs)

            raise ValueError(f"Unknown serializer: {serializer_name}")

        return method

    async def _decode(self, obj: t.Any, **kwargs: t.Dict[str, t.Any]) -> t.Any:
        if obj is None:
            raise ValueError("Cannot decode from None input")
        if isinstance(obj, (str, bytes)) and not obj:
            raise ValueError("Cannot decode from empty input")

        if isinstance(obj, AsyncPath):
            try:
                try:
                    obj = await obj.read_bytes()
                except (AttributeError, NotImplementedError):
                    text = await obj.read_text()
                    obj = text.encode("utf-8") if text else ""
            except FileNotFoundError:
                raise FileNotFoundError(f"File not found: {obj}")
        elif isinstance(obj, Path):
            try:
                try:
                    obj = obj.read_bytes()
                except (AttributeError, NotImplementedError):
                    text = obj.read_text()
                    obj = text.encode("utf-8") if text else ""
            except FileNotFoundError:
                raise FileNotFoundError(f"File not found: {obj}")

        if isinstance(obj, str) and self.serializer in (
            msgspec.json.decode,
            msgspec.yaml.decode,
            msgspec.toml.decode,
        ):
            obj = obj.encode("utf-8")

        try:
            return self.serializer(obj, **kwargs)
        except msgspec.DecodeError as e:
            raise msgspec.DecodeError(f"Failed to decode: {e}")  # type: ignore
        except toml.decoder.TomlDecodeError as e:
            raise toml.decoder.TomlDecodeError(f"Failed to decode: {e}")  # type: ignore
        except Exception as e:
            raise RuntimeError(f"Error during decoding: {e}") from e

    async def _encode(self, obj: t.Any, **kwargs: t.Dict[str, t.Any]) -> bytes:
        if isinstance(obj, (AsyncPath, Path)):
            if self.action == "decode":
                return await self._decode(obj, **kwargs)
            raise TypeError(f"Cannot encode a Path object directly: {obj}")

        obj = self._preprocess_data(obj, kwargs)

        data = self._serialize(obj, kwargs)

        if self.path is not None:
            await self._write_to_path(data)

        return data

    def _preprocess_data(self, obj: t.Any, kwargs: t.Dict[str, t.Any]) -> t.Any:
        if (
            self.serializer is msgspec.json.encode
            and self.sort_keys
            and isinstance(obj, dict)
        ):
            obj = {k: obj[k] for k in sorted(obj.keys())}

        if self.serializer is msgspec.yaml.encode:
            kwargs["sort_keys"] = self.sort_keys

        if self.serializer is msgspec.toml.encode and isinstance(obj, dict):
            for key, value in obj.items():
                if isinstance(value, list) and any(
                    isinstance(item, dict) for item in value
                ):
                    obj[key] = str(value)

        return obj

    def _serialize(self, obj: t.Any, kwargs: t.Dict[str, t.Any]) -> bytes:
        if self.serializer is msgspec.json.encode and kwargs.get("indent") is not None:
            indent = kwargs.pop("indent")
            return json.dumps(obj, indent=indent).encode()

        try:
            return self.serializer(obj, **kwargs)
        except TypeError as e:
            if (
                "Extra keyword arguments provided" in str(e)
                and self.serializer is msgspec.json.encode
            ):
                filtered_kwargs = {k: v for k, v in kwargs.items() if k != "indent"}
                return self.serializer(obj, **filtered_kwargs)
            raise

    async def _write_to_path(self, data: bytes) -> None:
        try:
            if isinstance(self.path, AsyncPath):
                await self.path.write_bytes(data)
            elif isinstance(self.path, Path):
                self.path.write_bytes(data)
        except PermissionError:
            raise PermissionError(f"Permission denied when writing to {self.path}")

    async def process(self, obj: t.Any, **kwargs: t.Dict[str, t.Any]) -> t.Any:
        if self.action in ("load", "decode"):
            return await self._decode(obj, **kwargs)
        if self.action in ("dump", "encode"):
            return await self._encode(obj, **kwargs)
        return None

    async def __call__(
        self,
        obj: t.Any,
        path: t.Optional[AsyncPath | Path] = None,
        sort_keys: bool = False,
        **kwargs: t.Dict[str, t.Any],
    ) -> t.Any:
        self.path = path
        self.sort_keys = sort_keys

        frame = sys._getframe(1)

        code_context = linecache.getline(frame.f_code.co_filename, frame.f_lineno)
        pattern = r"await\s+(\w+)\.(\w+)\("
        calling_method = search(pattern, code_context)

        if calling_method:
            caller_obj = calling_method.group(1)
            serializer_name = calling_method.group(2)

            if caller_obj in ("encode", "dump"):
                self.action = "encode"
            elif caller_obj in ("decode", "load"):
                self.action = "decode"
            else:
                self.action = "encode"

            if serializer_name in self.serializers:
                serializer = self.serializers[serializer_name]
                if self.action == "encode":
                    self.serializer = getattr(serializer, "encode")
                else:
                    self.serializer = getattr(serializer, "decode")

                return await self.process(obj, **kwargs)

        caller_name = frame.f_code.co_name
        if "encode" in caller_name or "dump" in caller_name:
            self.action = "encode"
        else:
            self.action = "decode"

        serializer = self.serializers["json"]
        if self.action == "encode":
            self.serializer = getattr(serializer, "encode")
        else:
            self.serializer = getattr(serializer, "decode")

        return await self.process(obj, **kwargs)


encode = Encode()
decode = Encode()
dump = Encode()
load = Encode()

for s in serializers.__dict__:
    setattr(decode, s, decode._create_method(s, "decode"))



"""Tests for the SecretManager adapter."""

import typing as t
from contextlib import asynccontextmanager, suppress
from unittest.mock import AsyncMock, MagicMock

import pytest
from google.api_core.exceptions import AlreadyExists, PermissionDenied
from google.cloud.secretmanager_v1 import (
    AccessSecretVersionRequest,
    AddSecretVersionRequest,
    CreateSecretRequest,
    DeleteSecretRequest,
    ListSecretsRequest,
    SecretManagerServiceAsyncClient,
)
from acb.config import Config


@pytest.fixture
def mock_async_context_manager() -> t.Callable[..., t.AsyncContextManager[MagicMock]]:
    @asynccontextmanager
    async def _async_context_manager(
        *args: t.Any, **kwargs: t.Any
    ) -> t.AsyncGenerator[MagicMock, None]:
        yield MagicMock()

    return _async_context_manager


class MockSecret:
    project: str = "test-project"
    parent: str = "projects/test-project"

    def __init__(self) -> None:
        self.config = MagicMock()
        self.logger = MagicMock()
        self.prefix = "test_app_"
        self._client = None

    def extract_secret_name(self, secret_path: str) -> t.Any:
        return secret_path.split("/")[-1].removeprefix(self.prefix)

    async def list(self, adapter: str) -> t.Any:
        request = ListSecretsRequest(
            parent=self.parent, filter=f"{self.prefix}{adapter}_"
        )
        try:
            client_secrets = await self.client.list_secrets(request=request)
        except PermissionDenied:
            raise SystemExit(
                "\n ERROR:  'project' id in 'settings/app.yml' is invalid or not set!\n"
            )
        client_secrets = [
            self.extract_secret_name(secret.name) async for secret in client_secrets
        ]
        return client_secrets

    async def get(self, name: str) -> t.Any:
        path = f"projects/{self.project}/secrets/{name}/versions/latest"
        request = AccessSecretVersionRequest(name=path)
        version = await self.client.access_secret_version(request=request)
        payload = version.payload.data.decode()
        self.logger.info(f"Fetched secret - {name}")
        return payload

    async def create(self, name: str, value: str) -> None:
        with suppress(AlreadyExists):
            request = CreateSecretRequest(
                parent=self.parent,
                secret_id=name,
                secret={"replication": {"automatic": {}}},
            )
            version = await self.client.create_secret(request)
            request = AddSecretVersionRequest(
                parent=version.name,
                payload={"data": value.encode()},
            )
            await self.client.add_secret_version(request)
            self.logger.debug(f"Created secret - {name}")

    async def update(self, name: str, value: str) -> None:
        secret = self.client.secret_path(self.project, name)
        request = AddSecretVersionRequest(
            parent=secret,
            payload={"data": value.encode()},
        )
        await self.client.add_secret_version(request)
        self.logger.debug(f"Updated secret - {name}")

    async def set(self, name: str, value: str) -> None:
        if await self.exists(name):
            await self.update(name, value)
        else:
            await self.create(name, value)

    async def exists(self, name: str) -> bool:
        try:
            await self.get(name)
            return True
        except Exception:
            return False

    async def delete(self, name: str) -> None:
        secret = self.client.secret_path(self.project, name)
        request = DeleteSecretRequest(name=secret)
        await self.client.delete_secret(request=request)
        self.logger.debug(f"Deleted secret - {secret}")

    @property
    def client(self) -> SecretManagerServiceAsyncClient:
        if self._client is None:
            self._client = SecretManagerServiceAsyncClient()
        return self._client

    async def init(self) -> None:
        pass

    async def list_versions(self, name: str) -> t.List[str]:
        return ["v1", "v2"]


class TestSecretManager:
    @pytest.fixture
    def mock_config(self) -> MagicMock:
        mock_config = MagicMock(spec=Config)
        mock_app = MagicMock()
        mock_app.name = "test_app"
        mock_config.app = mock_app
        return mock_config

    @pytest.fixture
    def mock_client(self) -> MagicMock:
        return MagicMock(spec=SecretManagerServiceAsyncClient)

    @pytest.fixture
    def secret_manager(
        self, mock_config: MagicMock, mock_client: MagicMock
    ) -> MockSecret:
        secret = MockSecret()
        secret.config = mock_config
        secret._client = mock_client
        return secret

    def test_extract_secret_name(self, secret_manager: MockSecret) -> None:
        project = "test-project"
        app_prefix = "test_app_"
        secret_name = "app_settings"

        secret_path = f"projects/{project}/secrets/{app_prefix}{secret_name}"

        result = secret_manager.extract_secret_name(secret_path)

        assert result == secret_name

    @pytest.mark.asyncio
    async def test_list(
        self, secret_manager: MockSecret, mock_client: MagicMock
    ) -> None:
        adapter = "test"
        mock_secret1 = MagicMock()
        mock_secret1.name = "projects/test-project/secrets/test_app_test_secret1"
        mock_secret2 = MagicMock()
        mock_secret2.name = "projects/test-project/secrets/test_app_test_secret2"

        mock_list_secrets = AsyncMock()
        mock_list_secrets.__aiter__.return_value = [mock_secret1, mock_secret2]
        mock_client.list_secrets.return_value = mock_list_secrets

        result = await secret_manager.list(adapter)

        mock_client.list_secrets.assert_called_once()
        assert len(result) == 2
        assert "test_secret1" in result
        assert "test_secret2" in result

    @pytest.mark.asyncio
    async def test_get(
        self, secret_manager: MockSecret, mock_client: MagicMock
    ) -> None:
        name = "test_secret"
        expected_value = "test_value"

        mock_version = MagicMock()
        mock_version.payload.data = expected_value.encode()
        mock_client.access_secret_version.return_value = mock_version

        result = await secret_manager.get(name)

        mock_client.access_secret_version.assert_called_once()
        assert result == expected_value

    @pytest.mark.asyncio
    async def test_create(
        self, secret_manager: MockSecret, mock_client: MagicMock
    ) -> None:
        name = "test_secret"
        value = "test_value"

        mock_version = MagicMock()
        mock_version.name = "test_version"
        mock_client.create_secret.return_value = mock_version

        await secret_manager.create(name, value)

        mock_client.create_secret.assert_called_once()
        mock_client.add_secret_version.assert_called_once()

    @pytest.mark.asyncio
    async def test_update(
        self, secret_manager: MockSecret, mock_client: MagicMock
    ) -> None:
        name = "test_secret"
        value = "test_value"
        mock_client.secret_path.return_value = f"projects/test-project/secrets/{name}"

        await secret_manager.update(name, value)

        mock_client.add_secret_version.assert_called_once()

    @pytest.mark.asyncio
    async def test_set_existing(self, secret_manager: MockSecret) -> None:
        name = "test_secret"
        value = "test_value"

        secret_manager.exists = AsyncMock(return_value=True)
        secret_manager.update = AsyncMock()
        secret_manager.create = AsyncMock()

        await secret_manager.set(name, value)

        secret_manager.exists.assert_called_once_with(name)
        secret_manager.update.assert_called_once_with(name, value)
        secret_manager.create.assert_not_called()

    @pytest.mark.asyncio
    async def test_set_new(self, secret_manager: MockSecret) -> None:
        name = "test_secret"
        value = "test_value"

        secret_manager.exists = AsyncMock(return_value=False)
        secret_manager.update = AsyncMock()
        secret_manager.create = AsyncMock()

        await secret_manager.set(name, value)

        secret_manager.exists.assert_called_once_with(name)
        secret_manager.create.assert_called_once_with(name, value)
        secret_manager.update.assert_not_called()

    @pytest.mark.asyncio
    async def test_exists_true(self, secret_manager: MockSecret) -> None:
        name = "test_secret"
        secret_manager.get = AsyncMock(return_value="test_value")

        result = await secret_manager.exists(name)

        assert result
        secret_manager.get.assert_called_once_with(name)

    @pytest.mark.asyncio
    async def test_exists_false(self, secret_manager: MockSecret) -> None:
        name = "test_secret"
        secret_manager.get = AsyncMock(side_effect=Exception("Not found"))

        result = await secret_manager.exists(name)

        assert not result
        secret_manager.get.assert_called_once_with(name)

    @pytest.mark.asyncio
    async def test_delete(
        self, secret_manager: MockSecret, mock_client: MagicMock
    ) -> None:
        name = "test_secret"
        secret_path = f"projects/test-project/secrets/{name}"
        mock_client.secret_path.return_value = secret_path

        await secret_manager.delete(name)

        mock_client.delete_secret.assert_called_once()

    @pytest.mark.asyncio
    async def test_list_versions(self, secret_manager: MockSecret) -> None:
        name = "test_secret"

        result = await secret_manager.list_versions(name)

        assert len(result) == 2
        assert "v1" in result
        assert "v2" in result

    @pytest.mark.asyncio
    async def test_init(self, secret_manager: MockSecret) -> None:
        await secret_manager.init()



"""Tests for the Infisical adapter."""

import typing as t
from contextlib import asynccontextmanager
from unittest.mock import AsyncMock, MagicMock

import pytest
from infisical_sdk import InfisicalSDKClient
from acb.config import Config


@pytest.fixture
def mock_async_context_manager() -> t.Callable[..., t.AsyncContextManager[MagicMock]]:
    @asynccontextmanager
    async def _async_context_manager(
        *args: t.Any, **kwargs: t.Any
    ) -> t.AsyncGenerator[MagicMock, None]:
        mock = MagicMock()
        mock.get_secret = MagicMock(return_value=MagicMock(secret_value="test-value"))
        yield mock

    return _async_context_manager


class MockInfisical:
    def __init__(self) -> None:
        self.config = MagicMock()
        self.logger = MagicMock()
        self.prefix = "test_app_"
        self._client = None
        self.project_id = "test-project-id"
        self.environment = "dev"
        self.client_cls = MagicMock()

    def extract_secret_name(self, secret_path: str) -> t.Any:
        return secret_path.split("/")[-1].removeprefix(self.prefix)

    async def list(self, adapter: str) -> t.List[str]:
        secrets = []
        async with self.client as client:
            client_secrets = client.secrets.list_secrets.return_value.secrets
            for secret in client_secrets:
                if secret.secretKey.startswith(f"{self.prefix}{adapter}_"):
                    secrets.append(self.extract_secret_name(secret.secretKey))
        return secrets

    async def get(self, name: str) -> str:
        async with self.client as client:
            secret = client.secrets.get_secret_by_name.return_value
            self.logger.info(f"Fetched secret - {name}")
            return secret.secretValue

    async def create(self, name: str, value: str) -> None:
        async with self.client as client:
            client.secrets.create_secret_by_name.return_value = None
            self.logger.debug(f"Created secret - {name}")

    async def update(self, name: str, value: str) -> None:
        async with self.client as client:
            client.secrets.update_secret_by_name.return_value = None
            self.logger.debug(f"Updated secret - {name}")

    async def set(self, name: str, value: str) -> None:
        if await self.exists(name):
            await self.update(name, value)
        else:
            await self.create(name, value)

    async def exists(self, name: str) -> bool:
        try:
            await self.get(name)
            return True
        except Exception:
            return False

    async def delete(self, name: str) -> None:
        async with self.client as client:
            client.secrets.delete_secret_by_name.return_value = None
            self.logger.debug(f"Deleted secret - {name}")

    async def list_versions(self, name: str) -> t.List[str]:
        return ["v1"]

    async def init(self) -> None:
        try:
            await self.list("test")
            self.logger.info("Infisical secret adapter initialized successfully")
        except Exception as e:
            self.logger.error(f"Failed to initialize Infisical secret adapter: {e}")
            raise

    @property
    def client(self) -> t.Any:
        if self._client is None:
            self._client = self.client_cls(
                client_id="test-client-id",
                client_secret="test-client-secret",
                org_id="test-org-id",
                project_id=self.project_id,
                environment=self.environment,
            )
        return self._client

    @client.setter
    def client(self, value: t.Any) -> None:
        self._client = value


class TestInfisical:
    @pytest.fixture
    def mock_config(self) -> MagicMock:
        mock_config = MagicMock(spec=Config)
        mock_app = MagicMock()
        mock_app.name = "test_app"
        mock_config.app = mock_app
        mock_config.deployed = True
        mock_infisical = MagicMock()
        mock_infisical.client_id = "test-client-id"
        mock_infisical.client_secret = "test-client-secret"
        mock_infisical.org_id = "test-org-id"
        mock_infisical.project_id = "test-project-id"
        mock_infisical.environment = "dev"
        mock_config.infisical = mock_infisical
        return mock_config

    @pytest.fixture
    def mock_client(self) -> MagicMock:
        client = MagicMock(spec=InfisicalSDKClient)

        secrets_mock = MagicMock()
        client.secrets = secrets_mock

        mock_secret_response = MagicMock()
        mock_secret_response.secretValue = "test-value"
        secrets_mock.get_secret_by_name.return_value = mock_secret_response

        mock_list_response = MagicMock()
        mock_secret1 = MagicMock()
        mock_secret1.secretKey = "test_app_test_secret1"
        mock_secret2 = MagicMock()
        mock_secret2.secretKey = "test_app_test_secret2"
        mock_list_response.secrets = [mock_secret1, mock_secret2]
        secrets_mock.list_secrets.return_value = mock_list_response

        secrets_mock.create_secret_by_name.return_value = None
        secrets_mock.update_secret_by_name.return_value = None
        secrets_mock.delete_secret_by_name.return_value = None

        return client

    @pytest.fixture
    def infisical(
        self,
        mock_config: MagicMock,
        mock_client: MagicMock,
        mock_async_context_manager: t.Callable[..., t.AsyncContextManager[MagicMock]],
    ) -> MockInfisical:
        infisical = MockInfisical()
        infisical.config = mock_config

        async_ctx = mock_async_context_manager()
        mock_enter = AsyncMock(return_value=mock_client)
        async_ctx.__aenter__ = mock_enter

        infisical._client = async_ctx

        return infisical

    def test_extract_secret_name(self, infisical: MagicMock) -> None:
        app_prefix = "test_app_"
        secret_name = "app_settings"
        secret_path = f"{app_prefix}{secret_name}"
        result = infisical.extract_secret_name(secret_path)
        assert result == secret_name

    @pytest.mark.asyncio
    async def test_list(self, infisical: MockInfisical, mock_client: MagicMock) -> None:
        adapter = "test_adapter"

        original_list = infisical.list

        async def mock_list(adapter: str) -> t.List[str]:
            assert adapter == adapter
            return ["secret1", "secret2"]

        infisical.list = mock_list

        result = await infisical.list(adapter)

        assert len(result) == 2
        assert "secret1" in result
        assert "secret2" in result

        infisical.list = original_list

    @pytest.mark.asyncio
    async def test_get(self, infisical: MockInfisical, mock_client: MagicMock) -> None:
        name = "test_secret"
        expected_value = "test-value"

        original_get = infisical.get

        async def mock_get(name: str) -> str:
            assert name == name
            return expected_value

        infisical.get = mock_get

        result = await infisical.get(name)

        assert result == expected_value

        infisical.get = original_get

    @pytest.mark.asyncio
    async def test_create(
        self, infisical: MockInfisical, mock_client: MagicMock
    ) -> None:
        name = "test_secret"
        value = "test_value"

        original_create = infisical.create
        create_called = False

        async def mock_create(name: str, value: str) -> None:
            nonlocal create_called
            assert name == name
            assert value == value
            create_called = True

        infisical.create = mock_create

        await infisical.create(name, value)

        assert create_called

        infisical.create = original_create

    @pytest.mark.asyncio
    async def test_update(
        self, infisical: MockInfisical, mock_client: MagicMock
    ) -> None:
        name = "test_secret"
        value = "test_value"

        original_update = infisical.update
        update_called = False

        async def mock_update(name: str, value: str) -> None:
            nonlocal update_called
            assert name == name
            assert value == value
            update_called = True

        infisical.update = mock_update

        await infisical.update(name, value)

        assert update_called

        infisical.update = original_update

    @pytest.mark.asyncio
    async def test_set_existing(self, infisical: MockInfisical) -> None:
        name = "test_secret"
        value = "test_value"

        infisical.exists = AsyncMock(return_value=True)
        infisical.update = AsyncMock()
        infisical.create = AsyncMock()

        await infisical.set(name, value)

        infisical.exists.assert_called_once_with(name)
        infisical.update.assert_called_once_with(name, value)
        infisical.create.assert_not_called()

    @pytest.mark.asyncio
    async def test_set_new(self, infisical: MockInfisical) -> None:
        name = "test_secret"
        value = "test_value"

        infisical.exists = AsyncMock(return_value=False)
        infisical.update = AsyncMock()
        infisical.create = AsyncMock()

        await infisical.set(name, value)

        infisical.exists.assert_called_once_with(name)
        infisical.create.assert_called_once_with(name, value)
        infisical.update.assert_not_called()

    @pytest.mark.asyncio
    async def test_exists_true(self, infisical: MockInfisical) -> None:
        name = "test_secret"
        infisical.get = AsyncMock(return_value="test_value")

        result = await infisical.exists(name)

        assert result
        infisical.get.assert_called_once_with(name)

    @pytest.mark.asyncio
    async def test_exists_false(self, infisical: MockInfisical) -> None:
        name = "test_secret"
        infisical.get = AsyncMock(side_effect=Exception("Not found"))

        result = await infisical.exists(name)

        assert not result
        infisical.get.assert_called_once_with(name)

    @pytest.mark.asyncio
    async def test_delete(
        self, infisical: MockInfisical, mock_client: MagicMock
    ) -> None:
        name = "test_secret"

        original_delete = infisical.delete
        delete_called = False

        async def mock_delete(name: str) -> None:
            nonlocal delete_called
            assert name == name
            delete_called = True

        infisical.delete = mock_delete

        await infisical.delete(name)

        assert delete_called

        infisical.delete = original_delete

    @pytest.mark.asyncio
    async def test_list_versions(self, infisical: MockInfisical) -> None:
        name = "test_secret"

        result = await infisical.list_versions(name)

        assert len(result) == 1
        assert "v1" in result

    @pytest.mark.asyncio
    async def test_init(self, infisical: MockInfisical, mock_client: MagicMock) -> None:
        original_list = infisical.list
        list_called = False

        async def mock_list(adapter: str) -> t.List[str]:
            nonlocal list_called
            list_called = True
            return ["secret1", "secret2"]

        infisical.list = mock_list

        await infisical.init()

        assert list_called
        assert infisical.logger.info.called

        infisical.list = original_list



"""Tests for the SQLModel adapter."""

import typing as t
from contextlib import suppress
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from anyio import Path as AsyncPath
from sqlmodel import Field, SQLModel
from acb.adapters.models.sqlmodel import ModelsSettings


class TestModel(SQLModel, table=True):  # type: ignore
    id: t.Optional[int] = Field(default=None, primary_key=True)
    name: str
    value: int


class TestModelsSettings:
    def test_init(self) -> None:
        settings = ModelsSettings()
        settings.path = AsyncPath("/test/models")  # type: ignore
        assert settings.path == AsyncPath("/test/models")


class TestModels:
    def __init__(self) -> None:
        self.config = MagicMock()
        self.logger = MagicMock()
        self.models = []
        self.settings: ModelsSettings | None = None
        self.sql = MagicMock()
        self._import_models_directory = AsyncMock()  # type: ignore
        self._import_models_file = AsyncMock()  # type: ignore

    @pytest.fixture
    def mock_config(self) -> MagicMock:
        mock_config: MagicMock = MagicMock()
        mock_config.get_path.return_value = "/path/to/models"
        return mock_config

    @pytest.fixture
    def models_settings(self) -> ModelsSettings:
        settings = ModelsSettings()
        return settings

    @pytest.fixture
    def models_instance(self, mock_config: MagicMock) -> "TestModels":
        instance = TestModels()
        instance.config = mock_config
        return instance

    async def import_models(self, path: AsyncPath) -> None:
        if await path.is_dir():
            await self._import_models_directory(path)
        else:
            await self._import_models_file(path)

    async def init(self, settings: ModelsSettings | None = None) -> None:
        self.settings = settings

    @pytest.mark.asyncio
    async def test_import_models_file(self, models_instance: "TestModels") -> None:
        path = AsyncPath("/path/to/models/model.py")
        with (
            patch.object(AsyncPath, "is_dir", new_callable=AsyncMock) as mock_is_dir,
            patch.object(AsyncPath, "is_file", new_callable=AsyncMock) as mock_is_file,
        ):
            mock_is_dir.return_value = False
            mock_is_file.return_value = True
            await models_instance.import_models(path)
            models_instance._import_models_file.assert_called_once_with(path)

    @pytest.mark.asyncio
    async def test_import_models_directory(
        self,
        models_instance: "TestModels",
    ) -> None:
        path = AsyncPath("/path/to/models")
        with patch.object(AsyncPath, "is_dir", new_callable=AsyncMock) as mock_is_dir:
            mock_is_dir.return_value = True
            await models_instance.import_models(path)
            models_instance._import_models_directory.assert_called_once_with(path)

    @pytest.mark.asyncio
    async def test_import_models_file_not_found(
        self, models_instance: "TestModels"
    ) -> None:
        path = AsyncPath("/path/to/models/model.py")
        with (
            patch.object(AsyncPath, "is_dir", new_callable=AsyncMock) as mock_is_dir,
            patch.object(AsyncPath, "is_file", new_callable=AsyncMock) as mock_is_file,
        ):
            mock_is_dir.return_value = False
            mock_is_file.return_value = True
            models_instance._import_models_file.side_effect = FileNotFoundError
            models_instance.logger = MagicMock()
            with suppress(FileNotFoundError):
                await models_instance.import_models(path)
            models_instance._import_models_file.assert_called_once_with(path)

    @pytest.mark.asyncio
    async def test_import_models_directory_not_found(
        self, models_instance: "TestModels"
    ) -> None:
        path = AsyncPath("/path/to/models")
        with patch.object(AsyncPath, "is_dir", new_callable=AsyncMock) as mock_is_dir:
            mock_is_dir.return_value = True
            models_instance._import_models_directory.side_effect = FileNotFoundError
            models_instance.logger = MagicMock()
            with suppress(FileNotFoundError):
                await models_instance.import_models(path)
            models_instance._import_models_directory.assert_called_once_with(path)

    @pytest.mark.asyncio
    async def test_init(
        self,
        models_instance: "TestModels",
    ) -> None:
        models_instance.import_models = AsyncMock()  # type: ignore

        settings = ModelsSettings()
        settings.path = AsyncPath("/path/to/models")  # type: ignore

        await models_instance.init(settings)

        assert models_instance.settings == settings

    def test_init_settings(self) -> None:
        settings = ModelsSettings()
        settings.path = AsyncPath("/path/to/models")  # type: ignore
        assert settings.path == AsyncPath("/path/to/models")



"""Tests for encoding functionality."""

import json
import pickle
import typing as t
from pathlib import Path
from unittest.mock import AsyncMock, patch

import msgpack
import msgspec
import pytest
import toml
import yaml
from acb.actions.encode import encode

TEST_DATA: t.Final[t.Dict[str, t.Any]] = {"name": "Test", "value": 123, "active": True}
NESTED_TEST_DATA: t.Final[t.Dict[str, t.Any]] = {
    "name": "Test",
    "value": 123,
    "active": True,
    "nested": {"key1": "value1", "key2": 234, "key3": False},
    "list": [1, 2, 3, "four", {"five": 5}],
}


class TestEncode:
    @pytest.fixture
    def async_tmp_path(self, tmp_path: Path) -> Path:
        return tmp_path

    @pytest.mark.asyncio
    async def test_json_encode(self) -> None:
        result_bytes: bytes = await encode.json(TEST_DATA)
        assert isinstance(result_bytes, bytes)
        decoded: t.Any = json.loads(result_bytes.decode())
        if isinstance(decoded, dict):
            dict_result: t.Dict[str, t.Any] = decoded
        else:
            dict_result: t.Dict[str, t.Any] = {}
        assert dict_result == TEST_DATA

    @pytest.mark.asyncio
    async def test_json_encode_with_indent(self) -> None:
        result_bytes: bytes = await encode.json(TEST_DATA, indent=4)
        assert isinstance(result_bytes, bytes)
        decoded: t.Any = json.loads(result_bytes.decode())
        if isinstance(decoded, dict):
            dict_result: t.Dict[str, t.Any] = decoded
        else:
            dict_result: t.Dict[str, t.Any] = {}
        assert dict_result == TEST_DATA
        lines: t.List[str] = result_bytes.decode().split("\n")
        assert len(lines) > 1

    @pytest.mark.asyncio
    async def test_json_encode_to_file(self, async_tmp_path: Path) -> None:
        file_path: Path = async_tmp_path / "test_data.json"

        with patch.object(Path, "write_bytes", AsyncMock()) as mock_write:
            await encode.json(TEST_DATA, path=file_path)

            assert mock_write.called
            call_args: bytes = mock_write.call_args[0][0]
            assert isinstance(call_args, bytes)
            decoded: t.Any = json.loads(call_args.decode())
            if isinstance(decoded, dict):
                dict_result: t.Dict[str, t.Any] = decoded
            else:
                dict_result: t.Dict[str, t.Any] = {}
            assert dict_result == TEST_DATA

    @pytest.mark.asyncio
    async def test_yaml_encode(self) -> None:
        result_bytes: bytes = await encode.yaml(TEST_DATA)

        assert isinstance(result_bytes, bytes)
        decoded: t.Any = yaml.safe_load(result_bytes.decode())
        if isinstance(decoded, dict):
            dict_result: t.Dict[str, t.Any] = decoded
        else:
            dict_result: t.Dict[str, t.Any] = {}
        assert dict_result == TEST_DATA

    @pytest.mark.asyncio
    async def test_yaml_encode_to_file(self, async_tmp_path: Path) -> None:
        file_path: Path = async_tmp_path / "test_data.yaml"

        with patch.object(Path, "write_bytes", AsyncMock()) as mock_write:
            await encode.yaml(TEST_DATA, path=file_path)

            assert mock_write.called
            call_args: bytes = mock_write.call_args[0][0]
            assert isinstance(call_args, bytes)
            decoded: t.Any = yaml.safe_load(call_args.decode())
            if isinstance(decoded, dict):
                dict_result: t.Dict[str, t.Any] = decoded
            else:
                dict_result: t.Dict[str, t.Any] = {}
            assert dict_result == TEST_DATA

    @pytest.mark.asyncio
    async def test_toml_encode(self) -> None:
        result_bytes: bytes = await encode.toml(TEST_DATA)

        assert isinstance(result_bytes, bytes)
        decoded: t.Any = toml.loads(result_bytes.decode())
        if isinstance(decoded, dict):
            dict_result: t.Dict[str, t.Any] = decoded
        else:
            dict_result: t.Dict[str, t.Any] = {}
        assert dict_result == TEST_DATA

    @pytest.mark.asyncio
    async def test_toml_encode_to_file(self, async_tmp_path: Path) -> None:
        file_path: Path = async_tmp_path / "test_data.toml"

        with patch.object(Path, "write_bytes", AsyncMock()) as mock_write:
            await encode.toml(TEST_DATA, path=file_path)

            assert mock_write.called
            call_args: bytes = mock_write.call_args[0][0]
            assert isinstance(call_args, bytes)
            decoded: t.Any = toml.loads(call_args.decode())
            if isinstance(decoded, dict):
                dict_result: t.Dict[str, t.Any] = decoded
            else:
                dict_result: t.Dict[str, t.Any] = {}
            assert dict_result == TEST_DATA

    @pytest.mark.asyncio
    async def test_msgpack_encode(self) -> None:
        result_bytes: bytes = await encode.msgpack(TEST_DATA)

        assert isinstance(result_bytes, bytes)
        decoded: t.Any = msgpack.unpackb(result_bytes, raw=False)
        if isinstance(decoded, dict):
            dict_result: t.Dict[str, t.Any] = decoded
        else:
            dict_result: t.Dict[str, t.Any] = {}
        assert dict_result == TEST_DATA

    @pytest.mark.asyncio
    async def test_msgpack_encode_to_file(self, async_tmp_path: Path) -> None:
        file_path: Path = async_tmp_path / "test_data.msgpack"

        with patch.object(Path, "write_bytes", AsyncMock()) as mock_write:
            await encode.msgpack(TEST_DATA, path=file_path)

            assert mock_write.called
            call_args: bytes = mock_write.call_args[0][0]
            assert isinstance(call_args, bytes)
            decoded: t.Any = msgpack.unpackb(call_args, raw=False)
            if isinstance(decoded, dict):
                dict_result: t.Dict[str, t.Any] = decoded
            else:
                dict_result: t.Dict[str, t.Any] = {}
            assert dict_result == TEST_DATA

    @pytest.mark.asyncio
    async def test_pickle_encode(self) -> None:
        result_bytes: bytes = await encode.pickle(TEST_DATA)

        assert isinstance(result_bytes, bytes)
        decoded: t.Any = pickle.loads(result_bytes)
        if isinstance(decoded, dict):
            dict_result: t.Dict[str, t.Any] = decoded
        else:
            dict_result: t.Dict[str, t.Any] = {}
        assert dict_result == TEST_DATA

    @pytest.mark.asyncio
    async def test_pickle_encode_to_file(self, async_tmp_path: Path) -> None:
        file_path: Path = async_tmp_path / "test_data.pickle"

        with patch.object(Path, "write_bytes", AsyncMock()) as mock_write:
            await encode.pickle(TEST_DATA, path=file_path)

            assert mock_write.called
            call_args: bytes = mock_write.call_args[0][0]
            assert isinstance(call_args, bytes)
            decoded: t.Any = pickle.loads(call_args)
            if isinstance(decoded, dict):
                dict_result: t.Dict[str, t.Any] = decoded
            else:
                dict_result: t.Dict[str, t.Any] = {}
            assert dict_result == TEST_DATA

    @pytest.mark.asyncio
    async def test_nested_data_json_encode(self) -> None:
        result_bytes: bytes = await encode.json(NESTED_TEST_DATA)

        assert isinstance(result_bytes, bytes)
        decoded: t.Any = json.loads(result_bytes.decode())
        if isinstance(decoded, dict):
            dict_result: t.Dict[str, t.Any] = decoded
        else:
            dict_result: t.Dict[str, t.Any] = {}
        assert dict_result == NESTED_TEST_DATA

    @pytest.mark.asyncio
    async def test_nested_data_yaml_encode(self) -> None:
        result_bytes: bytes = await encode.yaml(NESTED_TEST_DATA)

        assert isinstance(result_bytes, bytes)
        decoded: t.Any = yaml.safe_load(result_bytes.decode())
        if isinstance(decoded, dict):
            dict_result: t.Dict[str, t.Any] = decoded
        else:
            dict_result: t.Dict[str, t.Any] = {}
        assert dict_result == NESTED_TEST_DATA

    @pytest.mark.asyncio
    async def test_nested_data_toml_encode(self) -> None:
        result_bytes: bytes = await encode.toml(NESTED_TEST_DATA)

        assert isinstance(result_bytes, bytes)
        decoded: t.Any = toml.loads(result_bytes.decode())
        if isinstance(decoded, dict):
            dict_result: t.Dict[str, t.Any] = decoded
        else:
            dict_result: t.Dict[str, t.Any] = {}
        assert dict_result == NESTED_TEST_DATA

    @pytest.mark.asyncio
    async def test_nested_data_msgpack_encode(self) -> None:
        result_bytes: bytes = await encode.msgpack(NESTED_TEST_DATA)

        assert isinstance(result_bytes, bytes)
        decoded: t.Any = msgpack.unpackb(result_bytes, raw=False)
        if isinstance(decoded, dict):
            dict_result: t.Dict[str, t.Any] = decoded
        else:
            dict_result: t.Dict[str, t.Any] = {}
        assert dict_result == NESTED_TEST_DATA

    @pytest.mark.asyncio
    async def test_nested_data_pickle_encode(self) -> None:
        result_bytes: bytes = await encode.pickle(NESTED_TEST_DATA)

        assert isinstance(result_bytes, bytes)
        decoded: t.Any = pickle.loads(result_bytes)
        if isinstance(decoded, dict):
            dict_result: t.Dict[str, t.Any] = decoded
        else:
            dict_result: t.Dict[str, t.Any] = {}
        assert dict_result == NESTED_TEST_DATA

    @pytest.mark.asyncio
    async def test_json_encode_with_mock(self) -> None:
        with patch("json.dumps") as mock_dumps:
            mock_dumps.return_value = '{"mocked": true}'

            with patch.object(msgspec.json, "encode") as mock_encode:
                mock_encode.side_effect = lambda obj, **kwargs: mock_dumps(
                    obj, indent=kwargs.get("indent")
                ).encode()

                result_bytes: bytes = await encode.json(TEST_DATA)

                assert result_bytes == b'{"mocked": true}'
                mock_dumps.assert_called_once_with(TEST_DATA, indent=None)

    @pytest.mark.asyncio
    async def test_yaml_encode_with_mock(self) -> None:
        with patch("yaml.dump") as mock_dump:
            mock_dump.return_value = "mocked: true"

            with patch.object(msgspec.yaml, "encode") as mock_encode:
                mock_encode.side_effect = lambda obj, **kwargs: mock_dump(obj).encode()

                result_bytes: bytes = await encode.yaml(TEST_DATA)

                assert result_bytes == b"mocked: true"
                mock_dump.assert_called_once_with(TEST_DATA)

    @pytest.mark.asyncio
    async def test_toml_encode_with_mock(self) -> None:
        with patch("toml.dumps") as mock_dumps:
            mock_dumps.return_value = "mocked = true"

            with patch.object(msgspec.toml, "encode") as mock_encode:
                mock_encode.side_effect = lambda obj, **kwargs: mock_dumps(obj).encode()

                result_bytes: bytes = await encode.toml(TEST_DATA)

                assert result_bytes == b"mocked = true"
                mock_dumps.assert_called_once_with(TEST_DATA)

    @pytest.mark.asyncio
    async def test_msgpack_encode_with_mock(self) -> None:
        with patch("acb.actions.encode.Encode.process") as mock_process:
            mock_process.return_value = b"mocked"

            result_bytes: bytes = await encode.msgpack(TEST_DATA)

            assert result_bytes == b"mocked"

    @pytest.mark.asyncio
    async def test_pickle_encode_with_mock(self) -> None:
        with patch("acb.actions.encode.Encode.process") as mock_process:
            mock_process.return_value = b"mocked"

            result_bytes: bytes = await encode.pickle(TEST_DATA)

            assert result_bytes == b"mocked"

    @pytest.mark.asyncio
    async def test_pickle_encode_invalid(self) -> None:
        class UnpickleableObject:
            def __reduce__(self):
                raise TypeError("Cannot pickle this object")

        with pytest.raises(TypeError):
            await encode.pickle(UnpickleableObject())

    @pytest.mark.asyncio
    async def test_json_encode_invalid(self) -> None:
        with patch("json.dumps", side_effect=TypeError):
            with pytest.raises(TypeError):
                await encode.json(object())

    @pytest.mark.asyncio
    async def test_yaml_encode_invalid(self) -> None:
        with patch("yaml.dump", side_effect=TypeError):
            with pytest.raises(TypeError):
                await encode.yaml(object())

    @pytest.mark.asyncio
    async def test_toml_encode_invalid(self) -> None:
        with patch("toml.dumps", side_effect=TypeError):
            with pytest.raises(TypeError):
                await encode.toml(object())

    @pytest.mark.asyncio
    async def test_msgpack_encode_invalid(self) -> None:
        with patch("msgpack.packb", side_effect=TypeError):
            with pytest.raises(TypeError):
                await encode.msgpack(object())

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "data,expected_type",
        [
            (TEST_DATA, dict),
            (NESTED_TEST_DATA, dict),
        ],
    )
    async def test_json_encode_type_errors(
        self, data: t.Any, expected_type: type
    ) -> None:
        if not isinstance(data, dict):
            with pytest.raises(Exception):
                await encode.json(data)
        else:
            result_bytes: bytes = await encode.json(data)
            assert isinstance(result_bytes, bytes)
            decoded: t.Any = json.loads(result_bytes.decode())
            if isinstance(decoded, dict):
                dict_result: t.Dict[str, t.Any] = decoded
            else:
                dict_result: t.Dict[str, t.Any] = {}
            assert isinstance(dict_result, dict)

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "data,expected_type",
        [
            (TEST_DATA, dict),
            (NESTED_TEST_DATA, dict),
        ],
    )
    async def test_yaml_encode_type_errors(
        self, data: t.Any, expected_type: type
    ) -> None:
        if not isinstance(data, dict):
            with pytest.raises(Exception):
                await encode.yaml(data)
        else:
            result_bytes: bytes = await encode.yaml(data)
            assert isinstance(result_bytes, bytes)
            decoded: t.Any = yaml.safe_load(result_bytes.decode())
            if isinstance(decoded, dict):
                dict_result: t.Dict[str, t.Any] = decoded
            else:
                dict_result: t.Dict[str, t.Any] = {}
            assert isinstance(dict_result, dict)

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "data,expected_type",
        [
            (TEST_DATA, dict),
            (NESTED_TEST_DATA, dict),
        ],
    )
    async def test_toml_encode_type_errors(
        self, data: t.Any, expected_type: type
    ) -> None:
        if not isinstance(data, dict):
            with pytest.raises(Exception):
                await encode.toml(data)
        else:
            result_bytes: bytes = await encode.toml(data)
            assert isinstance(result_bytes, bytes)
            decoded: t.Any = toml.loads(result_bytes.decode())
            if isinstance(decoded, dict):
                dict_result: t.Dict[str, t.Any] = decoded
            else:
                dict_result: t.Dict[str, t.Any] = {}
            assert isinstance(dict_result, dict)

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "data,expected_type",
        [
            (TEST_DATA, dict),
            (NESTED_TEST_DATA, dict),
        ],
    )
    async def test_msgpack_encode_type_errors(
        self, data: t.Any, expected_type: type
    ) -> None:
        if not isinstance(data, dict):
            with pytest.raises(Exception):
                await encode.msgpack(data)
        else:
            result_bytes: bytes = await encode.msgpack(data)
            assert isinstance(result_bytes, bytes)
            decoded: t.Any = msgpack.unpackb(result_bytes, raw=False)
            if isinstance(decoded, dict):
                dict_result: t.Dict[str, t.Any] = decoded
            else:
                dict_result: t.Dict[str, t.Any] = {}
            assert isinstance(dict_result, dict)

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "data,expected_type",
        [
            (TEST_DATA, dict),
            (NESTED_TEST_DATA, dict),
        ],
    )
    async def test_pickle_encode_type_errors(
        self, data: t.Any, expected_type: type
    ) -> None:
        if not isinstance(data, dict):
            with pytest.raises(Exception):
                await encode.pickle(data)
        else:
            result_bytes: bytes = await encode.pickle(data)
            assert isinstance(result_bytes, bytes)
            decoded: t.Any = pickle.loads(result_bytes)
            if isinstance(decoded, dict):
                dict_result: t.Dict[str, t.Any] = decoded
            else:
                dict_result: t.Dict[str, t.Any] = {}
            assert isinstance(dict_result, dict)

    @pytest.mark.asyncio
    async def test_error_handling_file_operations(self, async_tmp_path: Path) -> None:
        test_file: Path = async_tmp_path / "test_data.json"

        with patch.object(Path, "write_bytes") as mock_write:
            mock_write.side_effect = PermissionError("Access denied")

            with pytest.raises(PermissionError):
                await encode.json(TEST_DATA, path=test_file)



"""Tests for the Firestore NoSQL adapter."""

from typing import Any, Callable, Optional, Type
from unittest.mock import MagicMock, PropertyMock, patch

import pytest
from google.cloud import firestore
from acb.adapters.nosql.firestore import Nosql as FirestoreNosql
from acb.adapters.nosql.firestore import NosqlSettings as FirestoreSettings


@pytest.fixture
async def mock_async_context_manager() -> Callable[[Optional[Any]], Any]:
    class MockAsyncContextManager:
        def __init__(self, mock_obj: Optional[Any] = None) -> None:
            self.mock_obj = mock_obj or MagicMock()

        async def __aenter__(self) -> Any:
            return self.mock_obj

        async def __aexit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_val: Optional[BaseException],
            exc_tb: Optional[Any],
        ) -> None:
            pass

    def _mock_async_context_manager(
        mock_obj: Optional[Any] = None,
    ) -> MockAsyncContextManager:
        return MockAsyncContextManager(mock_obj)

    return _mock_async_context_manager


class TestFirestoreSettings:
    def test_init(self) -> None:
        settings = FirestoreSettings(
            project_id="test-project",
            database="test-db",
            credentials_path="/path/to/credentials.json",
        )

        assert settings.project_id == "test-project"
        assert settings.database == "test-db"
        assert settings.credentials_path == "/path/to/credentials.json"

        settings = FirestoreSettings(
            project_id="test-project",
        )

        assert settings.project_id == "test-project"
        assert settings.database == "acb"
        assert settings.credentials_path is None


class TestFirestore:
    @pytest.fixture
    def nosql(self) -> FirestoreNosql:
        nosql = FirestoreNosql()
        nosql.config = MagicMock()
        nosql.logger = MagicMock()

        nosql.config.nosql.collection_prefix = ""

        return nosql

    def test_client_property(self, nosql: FirestoreNosql) -> None:
        mock_client = MagicMock(spec=firestore.Client)
        with patch.object(
            type(nosql), "client", new_callable=PropertyMock
        ) as mock_prop:
            mock_prop.return_value = mock_client
            assert nosql.client is not None
            assert nosql.client == mock_client

    def test_db_property(self, nosql: FirestoreNosql) -> None:
        mock_db = MagicMock()
        with patch.object(type(nosql), "db", new_callable=PropertyMock) as mock_prop:
            mock_prop.return_value = mock_db
            assert nosql.db is not None
            assert nosql.db == mock_db

    @pytest.mark.asyncio
    async def test_init(self, nosql: FirestoreNosql) -> None:
        original_init = nosql.init

        async def mock_init() -> None:
            nosql.logger.info("Initializing Firestore connection")

        nosql.init = mock_init

        await nosql.init()

        nosql.logger.info.assert_called_once()

        nosql.init = original_init

    @pytest.mark.asyncio
    async def test_init_error(self, nosql: FirestoreNosql) -> None:
        original_init = nosql.init

        async def mock_init():
            nosql.logger.info("Initializing Firestore connection")
            raise Exception("Connection error")

        nosql.init = mock_init

        with pytest.raises(Exception) as excinfo:
            await nosql.init()

        assert "Connection error" in str(excinfo.value)

        nosql.init = original_init

    def test_get_collection_ref(self, nosql: FirestoreNosql) -> None:
        original_method = nosql._get_collection_ref

        def mock_get_collection_ref(collection: str):
            prefix = nosql.config.nosql.collection_prefix
            mock = MagicMock()
            mock.name = f"{prefix}{collection}"
            return mock

        nosql._get_collection_ref = mock_get_collection_ref

        collection_ref = nosql._get_collection_ref("test_collection")

        assert collection_ref is not None
        assert collection_ref.name == "test_collection"

        nosql._get_collection_ref = original_method

    def test_convert_to_dict(self, nosql: FirestoreNosql) -> None:
        mock_doc = MagicMock()
        mock_doc.id = "test_id"
        mock_doc.to_dict.return_value = {"name": "test", "age": 30}

        result = nosql._convert_to_dict(mock_doc)

        assert result == {"_id": "test_id", "name": "test", "age": 30}



"""Tests for the Redis NoSQL adapter."""

from types import TracebackType
from typing import Any, List, Optional, Type
from unittest.mock import AsyncMock, MagicMock, PropertyMock

import pytest
from pydantic import SecretStr
from redis.asyncio import Redis
from acb.adapters.nosql.redis import Nosql as RedisNosql
from acb.adapters.nosql.redis import NosqlSettings as RedisSettings


@pytest.fixture
async def mock_async_context_manager(mock_obj: Optional[Any] = None):
    class MockAsyncContextManager:
        def __init__(self, mock_obj: Optional[Any] = None) -> None:
            self.mock_obj = mock_obj or MagicMock()

            async def async_execute() -> List[Any]:
                return []

            self.execute = async_execute

        async def __aenter__(self):
            return self.mock_obj

        async def __aexit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_val: Optional[BaseException],
            exc_tb: Optional[TracebackType],
        ) -> None:
            pass

    def _mock_async_context_manager(mock_obj: Optional[Any] = None):
        return MockAsyncContextManager(mock_obj)

    return _mock_async_context_manager


class TestRedisSettings:
    def test_init(self) -> None:
        settings = RedisSettings(
            host=SecretStr("localhost"),
            port=6379,
            db=0,
            cache_db=0,
            password=SecretStr("test_password"),
        )

        assert settings.host.get_secret_value() == "localhost"
        assert settings.port == 6379
        assert settings.db == 0
        assert settings.cache_db == 0
        assert settings.password.get_secret_value() == "test_password"

        settings = RedisSettings(
            host=SecretStr("localhost"),
        )

        assert settings.host.get_secret_value() == "localhost"
        assert settings.port == 6379
        assert settings.db == 0
        assert settings.cache_db == 0
        assert settings.password is None

    def test_cache_db_validator(self) -> None:
        with pytest.raises(ValueError):
            RedisSettings(
                host="localhost",
                database=0,
                cache_db=0,
            )


class TestRedis:
    @pytest.fixture
    def nosql(self, mock_async_context_manager: Any):
        nosql = RedisNosql()
        nosql.config = MagicMock()
        nosql.logger = MagicMock()

        nosql.config.nosql.collection_prefix = ""

        mock_client = MagicMock(spec=Redis)
        mock_client.get.return_value = b'{"_id": "test_id", "data": "test_data"}'
        mock_client.set.return_value = True
        mock_client.delete.return_value = 1
        mock_client.exists.return_value = 1
        mock_client.scan_iter.return_value = [b"collection:test_id"]

        pipeline_mock = MagicMock()

        async def async_execute() -> List[Any]:
            return []

        pipeline_mock.execute = async_execute
        mock_client.pipeline.return_value = pipeline_mock

        type(nosql).client = PropertyMock(return_value=mock_client)

        mock_om_client = MagicMock()
        type(nosql).om_client = PropertyMock(return_value=mock_om_client)

        return nosql

    def test_client_property(self, nosql: RedisNosql) -> None:
        assert nosql.client is not None

    def test_om_client_property(self, nosql: RedisNosql) -> None:
        assert nosql.om_client is not None

    @pytest.mark.asyncio
    async def test_init(self, nosql: RedisNosql) -> None:
        redis_settings = RedisSettings(
            host=SecretStr("localhost"),
            port=6379,
            db=0,
            connection_string="redis://localhost:6379/0",
        )

        nosql.config.nosql.get.return_value = redis_settings

        original_init = nosql.init

        async def mock_init_with_logging():
            nosql.logger.info("Initializing Redis connection")
            nosql.logger.info("Redis connection initialized successfully")
            return None

        mock_init = AsyncMock(side_effect=mock_init_with_logging)

        nosql.init = mock_init

        await nosql.init()
        assert nosql.logger.info.call_count == 2

        nosql.init = original_init

    @pytest.mark.asyncio
    async def test_init_error(self, nosql: RedisNosql) -> None:
        redis_settings = RedisSettings(
            host=SecretStr("localhost"),
            port=6379,
            db=0,
            connection_string="redis://localhost:6379/0",
        )

        nosql.config.nosql.get.return_value = redis_settings

        original_init = nosql.init

        mock_init_error = AsyncMock(side_effect=Exception("Connection error"))

        async def mock_init_with_logging():
            nosql.logger.info("Initializing Redis connection")
            nosql.logger.error(
                "Failed to initialize Redis connection: Connection error"
            )
            raise Exception("Connection error")

        mock_init_error.side_effect = mock_init_with_logging

        nosql.init = mock_init_error

        with pytest.raises(Exception) as excinfo:
            await nosql.init()
        assert "Connection error" in str(excinfo.value)
        assert nosql.logger.error.call_count == 1

        nosql.init = original_init

    @pytest.mark.asyncio
    async def test_get_key(self, nosql: RedisNosql) -> None:
        collection = "test_collection"
        id = "test_id"
        expected_key = f"{collection}:{id}"

        original_get_key = nosql._get_key

        mock_get_key = MagicMock(return_value=expected_key)

        nosql._get_key = mock_get_key

        result = nosql._get_key(collection, id)

        assert result == expected_key

        nosql._get_key = original_get_key

    @pytest.mark.asyncio
    async def test_matches_filter(self, nosql: RedisNosql) -> None:
        doc = {"name": "test", "value": 123}
        filter_dict = {"name": "test"}

        original_matches_filter = nosql._matches_filter

        mock_matches_filter = MagicMock(return_value=True)

        nosql._matches_filter = mock_matches_filter

        result = nosql._matches_filter(doc, filter_dict)

        assert result is True

        nosql._matches_filter = original_matches_filter

    @pytest.mark.asyncio
    async def test_transaction(self, nosql: RedisNosql) -> None:
        pipeline_mock = MagicMock()

        execute_mock = AsyncMock()

        pipeline_mock.execute = execute_mock
        pipeline_mock.discard = AsyncMock()

        client_mock = MagicMock()
        client_mock.pipeline.return_value = pipeline_mock

        original_client = nosql.client

        type(nosql).client = PropertyMock(return_value=client_mock)

        try:
            async with nosql.transaction():
                assert nosql._transaction == pipeline_mock

                nosql._transaction.set("test_key", "test_value")

            assert client_mock.pipeline.called

            assert execute_mock.called
        finally:
            type(nosql).client = PropertyMock(return_value=original_client)



"""Tests for the Cloudflare DNS adapter."""

import typing as t
from unittest.mock import AsyncMock, MagicMock

import pytest
from acb.adapters.dns.cloudflare import Dns as CloudflareDns


class MockClient:
    def __init__(self) -> None:
        self.zones = MagicMock()

        self.zones.list = MagicMock()
        self.zones.create = MagicMock()
        self.zones.get = MagicMock()


@pytest.fixture
def mock_client() -> MockClient:
    return MockClient()


@pytest.fixture
def cloudflare_dns(
    mock_client: MockClient,
) -> t.Any:
    dns = CloudflareDns()
    dns.name = "test-domain"  # type: ignore
    dns.subdomain = "www"  # type: ignore
    dns.domain = "example.com"  # type: ignore
    dns.email = "test@example.com"  # type: ignore
    dns.key = "cloudflare-api-key"  # type: ignore
    dns.config = MagicMock()
    dns.logger = MagicMock()

    dns.client = mock_client

    dns.get_zone_id = AsyncMock(return_value="zone-id")  # type: ignore
    dns._get_zone_id = AsyncMock()  # type: ignore
    dns.create_zone = MagicMock()  # type: ignore
    dns.list_records = MagicMock(return_value=[])  # type: ignore
    dns.create_records = MagicMock()  # type: ignore
    dns.delete_records = MagicMock()  # type: ignore
    dns.create_record = MagicMock()  # type: ignore

    return dns


class TestCloudflareDns:
    async def test_init(self, cloudflare_dns: t.Any) -> None:
        assert cloudflare_dns.name == "test-domain"  # type: ignore
        assert cloudflare_dns.subdomain == "www"  # type: ignore
        assert cloudflare_dns.domain == "example.com"  # type: ignore
        assert cloudflare_dns.email == "test@example.com"  # type: ignore
        assert cloudflare_dns.key == "cloudflare-api-key"  # type: ignore
        assert cloudflare_dns.zone_id is None

    @pytest.mark.asyncio
    async def test_get_zone_id(self, cloudflare_dns: t.Any) -> None:
        get_zone_id_mock: AsyncMock = cloudflare_dns.get_zone_id  # type: ignore
        get_zone_id_mock.reset_mock()
        get_zone_id_mock.return_value = "test-zone-id"

        result = await get_zone_id_mock()

        assert result == "test-zone-id"
        get_zone_id_mock.assert_called_once()

    @pytest.mark.asyncio
    async def test_get_zone_id_not_found(self, cloudflare_dns: t.Any) -> None:
        get_zone_id_mock: AsyncMock = cloudflare_dns.get_zone_id  # type: ignore
        get_zone_id_mock.reset_mock()
        get_zone_id_mock.side_effect = ValueError("Zone not found")

        with pytest.raises(ValueError, match="Zone not found"):
            await get_zone_id_mock()

        get_zone_id_mock.assert_called_once()
